//
// this is the histogramming module for the luminosity calculation done in the mezzanine card
//
module histLUM(
	input clk,					// this 40MHz clock is synchronized with the 80MHz clock
	input clkn,					// this 40MHz clock is 180 out of phase with above
	input resetn,				// active low reset
	input fEnable,				// global enable for the state machine to fill histograms
	input BC0,					// defined however you like it but 1 clock cycle long only please
	input zero,					// if set, then we zero the RAM entry when we read it out
	input rEnable,				// read enable for the state machine
	input [11:0] rAddress,	// address to read from (0 to 3653)
	input [3:0] rWhich,		// which of the 4 histograms do you want to read?
	input [5:0] sum0, sum1, sum2, sum3,	// here are the sums to histogram
	output reg[11:0] bucket,// LHC bucket number, 0 to 3653, synchronized to BC0 input
	output reg histDone,		// signals that we are finished with filling
	output reg goRead,		// signals that you can latch the data (when reading)
	output reg [15:0] hData0,hData1,hData2,hData3, // data presented to you when reading
	output latch_all,
	output orbitDone,
	output [3:0] latchEach,
	output reg enable_ram_read, enable_ram_write,
	output reg [2:0] state,
	// debug...
	output [15:0] rv1,rv2,rv3,rv4,
	output [1:0] rs1,rs2,rs3,rs4,
	output rw1,rw2,rw3,rw4,
	output [3:0] sel0,sel1,sel2,sel3,
	output [9:0] addrH0, addrH1,
	output [1:0] h0State, h1State,
	output reg [11:0] freeBucket,
	output szero		// spy on the zero wire

	);

//
// declarations...
//
wire latch0,latch1,latch2,latch3;		//these tell you that data is ready to latch (for reading histograms)
wire[15:0] hist0_value,hist1_value,hist2_value,hist3_value;//here are the histogram bin values
//reg enable_ram_write,enable_ram_read;
reg hist_zero;
assign szero = hist_zero;
//
// the following is for latching the output of each of the histograms which are selected for reading
//
wire whichAny = rWhich[0] | rWhich[1] | rWhich[2] | rWhich[3];
wire ready0 = rWhich[0] ? latch0 : 1;
wire ready1 = rWhich[1] ? latch1 : 1;
wire ready2 = rWhich[2] ? latch2 : 1;
wire ready3 = rWhich[3] ? latch3 : 1;
assign latch_all = ready0 & ready1 & ready2 & ready3 & whichAny;
assign latchEach = {ready3,ready2,ready1,ready0};
//
// we only allow 1365 orbits.  this is calculated from the fact that we have 16-bit bins
// (65535 max counts) and 48 channels, so that max increment for any bin is 48.  therefore,
// the max number of orbits is 65535/48=1365 so that there are no overflows
//
reg[10:0] hcounter;
wire hdone = (hcounter == 'd1364);
assign orbitDone = hdone;
//
// free running "bucket" that is reset with BC0, so we can see it whether we are filling or not
//
always @ (posedge clk) 
	if (~resetn) freeBucket <= 0;
	else freeBucket <= BC0 ? 0 : freeBucket + 1;
//
// here is the state machine, typical overkill but it works!
//
//reg[2:0] state;
parameter[2:0] WAIT=0, WAIT_BC0=1, RUN=2, DONE=3, READ=4, LATCH=5, WAIT_END=6;
always @ (posedge clk) begin
  if (~resetn) begin
		state <= WAIT;
		histDone <= 0;
		enable_ram_write <= 0;
		enable_ram_read <= 0;
		hcounter <= 0;
		bucket <= 0;
		hist_zero <= 0;
		hData0 <= 0;
		hData1 <= 0;
		hData2 <= 0;
		hData3 <= 0;
		goRead <= 0;
	end
  else case (state)
  	  WAIT: begin
			//
			// wait for either histogramming to begin, or reading histogramming to begin
			//
	  		histDone <= 0;
			goRead <= 0;
			enable_ram_write <= 0;
			enable_ram_read <= 0;
			hist_zero <= 0;
			bucket <= 0;
//			hData0 <= 0;
//			hData1 <= 0;
//			hData2 <= 0;
//			hData3 <= 0;
			hcounter <= 1;
			case ({fEnable,rEnable})
				2'b00:	state <= WAIT;
				2'b01:   state <= READ;
				2'b10:   state <= WAIT_BC0;
				2'b11:   state <= WAIT_BC0;  // filling gets priority, but...shouldn't happen
			endcase
			end
	   WAIT_BC0: begin
			//
			// wait for the next bc0 to come around before starting, so that
			// we can start filling at the beginning of the orbit
			//
			if (BC0) begin
			    //
				// set the ram enable so that it comes up when bucket is still at 0.
				// otherwise, we will miss the first bucket=0 bin!
				//
		      enable_ram_write <= 1;
				state <= RUN;
				end
			else state <= WAIT_BC0;
			end
	   RUN: begin
			//
			// ok, increment the bucket number and start filling.  wait for 1365 orbits to stop
			//
			bucket <= BC0 ? 0 : bucket + 1;
			if (BC0) hcounter <= hcounter + 1;
			if (hdone || ~fEnable) state <= DONE;
			else state <= RUN;
			end
	   DONE: begin
			//
			// wait for fEnable to go away before returning to the WAIT state
			//
			histDone <= 1;
			enable_ram_write <= 0;
			if (fEnable) state <= DONE;
			else state <= WAIT;
			end
	   READ: begin
			//
			// here is where we begin the read.  first assert the enable and the bin number
			// and wait for all the data to appear
			//
			bucket <= rAddress;
			enable_ram_read <= 1;
			if (latch_all) state <= LATCH;
			else state <= READ;
			end
	   LATCH: begin
			//
			// latch the 4 values here
			//
			hData0 <= hist0_value;
			hData1 <= hist1_value;
			hData2 <= hist2_value;
			hData3 <= hist3_value;
			goRead <= 1;
			state <= WAIT_END;
			end
	   WAIT_END: begin
			//
			// set zero according to what is in the writeonly register and wait for the localbus
			// read to go away
			//
			hist_zero <= zero;
			if (rEnable) state <= WAIT_END;
			else state <= WAIT;
		   end
  endcase
end

wire read0 = enable_ram_read & rWhich[0];
wire read1 = enable_ram_read & rWhich[1];
wire read2 = enable_ram_read & rWhich[2];
wire read3 = enable_ram_read & rWhich[3];
//wire[3:0] sel0,sel1,sel2,sel3;
wire r0,r1,r2,r3;
wire[1:0] hstate0,hstate1,hstate2,hstate3;
assign h0State = hstate0;
assign h1State = hstate1;
wire[9:0] addr0,addr1,addr2,addr3;
assign addrH0 = addr0;
assign addrH1 = addr1;
hbunch histo0 (.clock(clk),.nclock(clkn),.resetn(resetn),.bin(bucket),
	.enable_ram_write(enable_ram_write),.enable_ram_read(read0),.zero(hist_zero),
	.select(sel0),.readit(r0),.hstate(hstate0),.addrin(addr0),
	.rv1(),.rv2(),.rv3(),.rv4(),.rstate1(),.rstate2(),.rstate3(),.rstate4(),
	.rwrite1(),.rwrite2(),.rwrite3(),.rwrite4(),
	.weight(sum0),.latch_value(latch0),.value(hist0_value));   // these change for each histogram
//
// debug this one (initially)...
//
///wire [15:0] rv1,rv2,rv3,rv4;
///wire [1:0] rs1,rs2,rs3,rs4;
///wire rw1,rw2,rw3,rw4;
hbunch histo1 (.clock(clk),.nclock(clkn),.resetn(resetn),.bin(bucket),
	.enable_ram_write(enable_ram_write),.enable_ram_read(read1),.zero(hist_zero),
	.select(sel1),.readit(r1),.hstate(hstate1),.addrin(addr1),
//	.rv1(),.rv2(),.rv3(),.rv4(),.rstate1(),.rstate2(),.rstate3(),.rstate4(),
//	.rwrite1(),.rwrite2(),.rwrite3(),.rwrite4(),
	.rv1(rv1),.rv2(rv2),.rv3(rv3),.rv4(rv4),
	.rstate1(rs1),.rstate2(rs2),.rstate3(rs3),.rstate4(rs4),
	.rwrite1(rw1),.rwrite2(rw2),.rwrite3(rw3),.rwrite4(rw4),
	.weight(sum1),.latch_value(latch1),.value(hist1_value));   // these change for each histogram
//
hbunch histo2 (.clock(clk),.nclock(clkn),.resetn(resetn),.bin(bucket),
	.enable_ram_write(enable_ram_write),.enable_ram_read(read2),.zero(hist_zero),
	.select(sel2),.readit(r2),.hstate(hstate2),.addrin(addr2),
//	.rv1(rv1),.rv2(rv2),.rv3(rv3),.rv4(rv4),
//	.rstate1(rs1),.rstate2(rs2),.rstate3(rs3),.rstate4(rs4),
//	.rwrite1(rw1),.rwrite2(rw2),.rwrite3(rw3),.rwrite4(rw4),
	.rv1(),.rv2(),.rv3(),.rv4(),.rstate1(),.rstate2(),.rstate3(),.rstate4(),
	.rwrite1(),.rwrite2(),.rwrite3(),.rwrite4(),
	.weight(sum2),.latch_value(latch2),.value(hist2_value));   // these change for each histogram
hbunch histo3 (.clock(clk),.nclock(clkn),.resetn(resetn),.bin(bucket),
	.enable_ram_write(enable_ram_write),.enable_ram_read(read3),.zero(hist_zero),
	.select(sel3),.readit(r3),.hstate(hstate3),.addrin(addr3),
	.rv1(),.rv2(),.rv3(),.rv4(),.rstate1(),.rstate2(),.rstate3(),.rstate4(),
	.rwrite1(),.rwrite2(),.rwrite3(),.rwrite4(),
	.weight(sum3),.latch_value(latch3),.value(hist3_value));   // these change for each histogram

endmodule